@@ -60,7 +60,6 @@ readonly CONFIG_TRACING_CONFIG="test/config/config-tracing.yaml"
 # Strimzi Kafka Cluster Brokers URL (base64 encoded value for k8s secret)
 readonly STRIMZI_KAFKA_NAMESPACE="kafka" # Installation Namespace
 readonly STRIMZI_KAFKA_CLUSTER_BROKERS="my-cluster-kafka-bootstrap.kafka.svc:9092"
-readonly STRIMZI_KAFKA_CLUSTER_BROKERS_ENCODED="bXktY2x1c3Rlci1rYWZrYS1ib290c3RyYXAua2Fma2Euc3ZjOjkwOTI=" # Simple base64 encoding of STRIMZI_KAFKA_CLUSTER_BROKERS
 
 # Eventing Kafka main config path from HEAD.
 readonly KAFKA_CRD_CONFIG_TEMPLATE_DIR="./config/channel"
@@ -74,17 +73,17 @@ readonly EVENTING_KAFKA_SECRET_TEMPLATE="300-kafka-secret.yaml"
 readonly EVENTING_KAFKA_CONFIG_TEMPLATE="300-eventing-kafka-configmap.yaml"
 
 # Strimzi installation config template used for starting up Kafka clusters.
-readonly STRIMZI_INSTALLATION_CONFIG_TEMPLATE="test/config/100-strimzi-cluster-operator-0.20.0.yaml"
+readonly STRIMZI_INSTALLATION_CONFIG_TEMPLATE="test/config/100-strimzi-cluster-operator-0.23.0.yaml"
 # Strimzi installation config.
 readonly STRIMZI_INSTALLATION_CONFIG="$(mktemp "${ARTIFACTS}/strimzi-XXXXX.yaml")"
 # Kafka cluster CR config file.
-readonly KAFKA_INSTALLATION_CONFIG="test/config/100-kafka-ephemeral-triple-2.6.0.yaml"
+readonly KAFKA_INSTALLATION_CONFIG="test/config/100-kafka-ephemeral-triple-2.8.0.yaml"
 # Kafka TLS ConfigMap.
 readonly KAFKA_TLS_CONFIG="test/config/config-kafka-tls.yaml"
 # Kafka SASL ConfigMap.
 readonly KAFKA_SASL_CONFIG="test/config/config-kafka-sasl.yaml"
 # Kafka Users CR config file.
-readonly KAFKA_USERS_CONFIG="test/config/100-strimzi-users-0.20.0.yaml"
+readonly KAFKA_USERS_CONFIG="test/config/100-strimzi-users-0.23.0.yaml"
 # Kafka PLAIN cluster URL
 readonly KAFKA_PLAIN_CLUSTER_URL="my-cluster-kafka-bootstrap.kafka.svc.cluster.local:9092"
 # Kafka TLS cluster URL
@@ -102,9 +101,13 @@ readonly KAFKA_CRD_CONFIG_DIR="$(mktemp -d "${ARTIFACTS}/channel-crd-XXXXX")"
 # Real Kafka Source CRD config, generated from the template directory and
 # modified template file.
 readonly KAFKA_SOURCE_CRD_CONFIG_DIR="$(mktemp -d "${ARTIFACTS}/source-crd-XXXXX")"
+# A target directory where post install scripts are gets created.
+readonly KAFKA_POST_INSTALL_DIR="$(mktemp -d "${ARTIFACTS}/post-install-XXXXX")"
+readonly KAFKA_POST_INSTALL_TEMPLATE_DIR="config/post-install"
 
-# Kafka channel CRD config template directory.
-readonly KAFKA_SOURCE_TEMPLATE_DIR="config/source"
+# Kafka ST and MT Source CRD config template directory
+readonly KAFKA_SOURCE_TEMPLATE_DIR="config/source/single"
+readonly KAFKA_MT_SOURCE_TEMPLATE_DIR="config/source/multi"
 
 # Namespaces where we install Eventing components
 # This is the namespace of knative-eventing itself
@@ -181,10 +184,10 @@ function install_knative_eventing {
     # Install MT Channel Based Broker
     ko apply -f "${EVENTING_MT_CHANNEL_BROKER_CONFIG}"
     # Install IMC
-    ko apply -f "${EVENTING_IN_MEMORY_CHANNEL_CONFIG}"
+    ko apply -Rf "${EVENTING_IN_MEMORY_CHANNEL_CONFIG}"
     popd
   fi
-  wait_until_pods_running "${EVENTING_NAMESPACE}" || fail_test "Knative Eventing did not come up"
+   wait_until_pods_running "${EVENTING_NAMESPACE}" || fail_test "Knative Eventing did not come up"
 
   install_zipkin
 }
@@ -216,7 +219,7 @@ function knative_teardown() {
     pushd .
     cd ${GOPATH}/src/knative.dev/eventing
     # Remove IMC
-    ko delete -f "${EVENTING_IN_MEMORY_CHANNEL_CONFIG}"
+    ko delete -Rf "${EVENTING_IN_MEMORY_CHANNEL_CONFIG}"
     # Remove MT Channel Based Broker
     ko delete -f "${EVENTING_MT_CHANNEL_BROKER_CONFIG}"
     # Remove eventing
@@ -289,6 +292,7 @@ function install_consolidated_channel_crds {
     sed -i "s/REPLACE_WITH_CLUSTER_URL/${KAFKA_CLUSTER_URL}/" \
       "${KAFKA_CRD_CONFIG_DIR}/${KAFKA_CRD_CONFIG_TEMPLATE}"
     ko apply -f "${KAFKA_CRD_CONFIG_DIR}"
+    run_postinstall_jobs
   elif [[ "${source}" == 'latest-release' ]]; then
     ver="${LATEST_RELEASE_VERSION}"
     echo "Installing consolidated Kafka Channel CRD (from latest release: ${ver})"
@@ -345,14 +349,40 @@ function install_consolidated_sources_crds() {
     return 56
   fi
 
+  run_postinstall_jobs
   wait_until_pods_running "${EVENTING_NAMESPACE}" || fail_test "Failed to install the consolidated Kafka Source CRD"
 }
 
+function run_postinstall_jobs() {
+  # There are no post-install scripts today. This needs to be enabled if any post-install script is added
+
+  #  echo "Running post-install jobs using ${KAFKA_POST_INSTALL_DIR}"
+  #  rm -rf "${KAFKA_POST_INSTALL_DIR}" && mkdir -p "${KAFKA_POST_INSTALL_DIR}"
+  #  cp "${KAFKA_POST_INSTALL_TEMPLATE_DIR}/"*yaml "${KAFKA_POST_INSTALL_DIR}"
+  #  sed -i "s/namespace: knative-eventing/namespace: ${SYSTEM_NAMESPACE}/g" \
+  #    "${KAFKA_POST_INSTALL_DIR}/"*yaml
+  #  ko apply -f "${KAFKA_POST_INSTALL_DIR}"
+
+  echo "No postinstall jobs to run"
+}
+
 # Uninstall The eventing-kafka KafkaChannel Implementation Via Ko
 function uninstall_channel_crds() {
   echo "Uninstalling Kafka Channel CRD"
   kubectl delete secret -n "${SYSTEM_NAMESPACE}" kafka-cluster
-  sleep 10 # Give Controller Time To React To Kafka Secret Deletion ; )
+
+  # The distributed channel controller no longer actively monitors the kafka-cluster secret,
+  # so the receiver will only be torn down if there is a change to the kafkachannels or the
+  # config-kafka configmap.  However, if the secret is missing, the global resync does not happen
+  # (the controller presumes this missing secret is an error and reacts accordingly), so this
+  # manual deletion of the receiver deployment is the easiest solution here.
+  if [[ $1 == "distributed" ]]; then
+    # Create the same value as GenerateHash() in distributed/controller/util/hash.go
+    hash=$(md5 -qs "kafka-cluster" | cut -c 1-8)
+    kubectl delete deployment -n "${SYSTEM_NAMESPACE}" kafka-cluster-${hash}-receiver
+    kubectl delete service -n "${SYSTEM_NAMESPACE}" kafka-cluster-${hash}-receiver
+  fi
+
   echo "Current namespaces:"
   kubectl get namespaces
   echo "Current kafkachannels:"
@@ -371,8 +401,8 @@ function install_distributed_channel_crds() {
   cp "${DISTRIBUTED_TEMPLATE_DIR}/"*yaml "${KAFKA_CRD_CONFIG_DIR}"
   sed -i "s/namespace: knative-eventing/namespace: ${SYSTEM_NAMESPACE}/g" "${KAFKA_CRD_CONFIG_DIR}/"*yaml
 
-  # Update The Kafka Secret With Strimzi Kafka Cluster Brokers (No Authentication)
-  sed -i "s/brokers: \"\"/brokers: ${STRIMZI_KAFKA_CLUSTER_BROKERS_ENCODED}/" "${KAFKA_CRD_CONFIG_DIR}/${EVENTING_KAFKA_SECRET_TEMPLATE}"
+  # Update The ConfigMap With Strimzi Kafka Cluster Brokers (No Authentication)
+  sed -i "s/REPLACE_WITH_CLUSTER_URL/${KAFKA_CLUSTER_URL}/" ${KAFKA_CRD_CONFIG_DIR}/${EVENTING_KAFKA_CONFIG_TEMPLATE}
 
   # Update the config-kafka configmap to disable SASL/TLS for the tests
   sed -i '/^ *TLS:/{n;s/true/false/};/^ *SASL:/{n;s/true/false/}' "${KAFKA_CRD_CONFIG_DIR}/${EVENTING_KAFKA_CONFIG_TEMPLATE}"
@@ -383,9 +413,29 @@ function install_distributed_channel_crds() {
    # Add The kn-eventing-test-pull-secret (If Present) To ServiceAccount & Restart eventing-kafka Deployment
   add_kn_eventing_test_pull_secret "${SYSTEM_NAMESPACE}" eventing-kafka-channel-controller eventing-kafka-channel-controller
 
+  run_postinstall_jobs
+
   wait_until_pods_running "${SYSTEM_NAMESPACE}" || fail_test "Failed to install the distributed Kafka Channel CRD"
 }
 
+function install_mt_source() {
+  echo "Installing multi-tenant Kafka Source"
+  rm -rf "${KAFKA_SOURCE_CRD_CONFIG_DIR}" && mkdir -p "${KAFKA_SOURCE_CRD_CONFIG_DIR}"
+  cp "${KAFKA_MT_SOURCE_TEMPLATE_DIR}/"*yaml "${KAFKA_SOURCE_CRD_CONFIG_DIR}"
+  sed -i "s/namespace: knative-eventing/namespace: ${SYSTEM_NAMESPACE}/g" "${KAFKA_SOURCE_CRD_CONFIG_DIR}/"*yaml
+  ko apply -f "${KAFKA_SOURCE_CRD_CONFIG_DIR}" || return 1
+
+  # FIXME: integration-test-mt-source fails if post-install scripts are executed (knative-sandbox/eventing-kafka#495)
+  # run_postinstall_jobs
+
+  wait_until_pods_running "${EVENTING_NAMESPACE}" || fail_test "Failed to install the multi-tenant Kafka Source"
+}
+
+function uninstall_mt_source() {
+  echo "Uninstalling  multi-tenant Kafka Source CRD"
+  ko delete --ignore-not-found=true --now --timeout 180s -f "${KAFKA_SOURCE_CRD_CONFIG_DIR}"
+}
+
 function kafka_setup() {
   # Create The Namespace Where Strimzi Kafka Will Be Installed
   echo "Installing Kafka Cluster"
@@ -395,6 +445,7 @@ function kafka_setup() {
   sed "s/namespace: .*/namespace: ${STRIMZI_KAFKA_NAMESPACE}/" ${STRIMZI_INSTALLATION_CONFIG_TEMPLATE} > "${STRIMZI_INSTALLATION_CONFIG}"
 
   echo "Create The Actual Kafka Cluster Instance For The Cluster Operator To Setup using: ${STRIMZI_INSTALLATION_CONFIG}"
+  kubectl apply -f "${STRIMZI_INSTALLATION_CONFIG}" -n "${STRIMZI_KAFKA_NAMESPACE}" -l strimzi.io/crd-install=true
   kubectl apply -f "${STRIMZI_INSTALLATION_CONFIG}" -n "${STRIMZI_KAFKA_NAMESPACE}"
   kubectl apply -f "${KAFKA_INSTALLATION_CONFIG}" -n "${STRIMZI_KAFKA_NAMESPACE}"
 
@@ -455,8 +506,8 @@ function test_consolidated_channel_plain() {
   install_consolidated_channel_crds || return 1
   install_consolidated_sources_crds || return 1
 
-  go_test_e2e -tags=e2e,source -timeout=40m -test.parallel=${TEST_PARALLEL} ./test/e2e -channels=messaging.knative.dev/v1alpha1:KafkaChannel,messaging.knative.dev/v1beta1:KafkaChannel  || fail_test
-  go_test_e2e -tags=e2e,source -timeout=5m -test.parallel=${TEST_PARALLEL} ./test/conformance -channels=messaging.knative.dev/v1beta1:KafkaChannel -sources=sources.knative.dev/v1beta1:KafkaSource || fail_test
+  go_test_e2e -tags=e2e,consolidated,source -timeout=40m -test.parallel=${TEST_PARALLEL} ./test/e2e -channels=messaging.knative.dev/v1beta1:KafkaChannel  || fail_test
+  go_test_e2e -tags=e2e,consolidated,source -timeout=5m -test.parallel=${TEST_PARALLEL} ./test/conformance -channels=messaging.knative.dev/v1beta1:KafkaChannel -sources=sources.knative.dev/v1beta1:KafkaSource || fail_test
 
   uninstall_sources_crds || return 1
   uninstall_channel_crds || return 1
@@ -472,7 +523,7 @@ function test_consolidated_channel_tls() {
 
   install_consolidated_channel_crds || return 1
 
-  go_test_e2e -tags=e2e -timeout=40m -test.parallel=${TEST_PARALLEL} ./test/e2e -channels=messaging.knative.dev/v1beta1:KafkaChannel  || fail_test
+  go_test_e2e -tags=e2e,consolidated -timeout=40m -test.parallel=${TEST_PARALLEL} ./test/e2e -channels=messaging.knative.dev/v1beta1:KafkaChannel  || fail_test
 
   uninstall_channel_crds || return 1
 }
@@ -487,7 +538,7 @@ function test_consolidated_channel_sasl() {
 
   install_consolidated_channel_crds || return 1
 
-  go_test_e2e -tags=e2e -timeout=40m -test.parallel=${TEST_PARALLEL} ./test/e2e -channels=messaging.knative.dev/v1beta1:KafkaChannel  || fail_test
+  go_test_e2e -tags=e2e,consolidated -timeout=40m -test.parallel=${TEST_PARALLEL} ./test/e2e -channels=messaging.knative.dev/v1beta1:KafkaChannel  || fail_test
 
   uninstall_channel_crds || return 1
 }
@@ -502,9 +553,39 @@ function test_distributed_channel() {
   go_test_e2e -tags=e2e -timeout=40m -test.parallel=${TEST_PARALLEL} ./test/e2e -channels=messaging.knative.dev/v1beta1:KafkaChannel  || fail_test
   go_test_e2e -tags=e2e -timeout=5m -test.parallel=${TEST_PARALLEL} ./test/conformance -channels=messaging.knative.dev/v1beta1:KafkaChannel || fail_test
 
-  uninstall_channel_crds || return 1
+  uninstall_channel_crds distributed || return 1
 }
 
+# Installs the resources necessary to test the multi-tenant source, runs those tests, and then cleans up those resources
+function test_mt_source() {
+  echo "Testing the multi-tenant source"
+  install_mt_source || return 1
+
+  export TEST_MT_SOURCE
+
+  echo "Run rekt tests"
+  go_test_e2e -tags=e2e -timeout=20m -test.parallel=${TEST_PARALLEL} ./test/rekt/... || fail_test
+
+  # still run those since some test cases are still missing
+  echo "Run classic tests"
+  go_test_e2e -tags=source,mtsource -timeout=20m -test.parallel=${TEST_PARALLEL} ./test/e2e/...  || fail_test
+
+  # wait for all KafkaSources to be deleted
+  local iterations=0
+  local progress="Waiting for KafkaSources to be deleted..."
+  while [[ "$(kubectl get kafkasources --all-namespaces)" != "" && $iterations -lt 60 ]]
+  do
+    echo -ne "${progress}\r"
+    progress="${progress}."
+    iterations=$((iterations + 1))
+    kubectl get kafkasources --all-namespaces -oyaml
+    sleep 3
+  done
+
+  uninstall_mt_source || return 1
+}
+
+
 function parse_flags() {
   # This function will be called repeatedly by initialize() with one fewer
   # argument each time and expects a return value of "the number of arguments to skip"
@@ -527,6 +608,10 @@ function parse_flags() {
       TEST_CONSOLIDATED_CHANNEL_SASL=1
       return 1
       ;;
+    --mt-source)
+      TEST_MT_SOURCE=1
+      return 1
+      ;;
   esac
   return 0
 }
